/*********************************************************************************
 * dtmfbox.c ($Revision: 261 $)
 *
 * Copyright (c) 2006-2009 Marco Zissen <maz@v3v.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *********************************************************************************/
#include "global_headers.h"
#include "global_objects.h"

#define THIS_FILE "dtmfbox.c"
#define __DTMFBOX_C

#if DTMFBOX_WIN32
    HANDLE hMutex, hWnd;
    #define SIGHUP 1
#endif

// Start application as daemon (fork process)
int daemonize_app()
{   
    // Windows fork ;)
    #if DTMFBOX_WIN32        
        HANDLE hConsole = NULL;
        hConsole = FindWindowEx(hWnd, NULL, "ConsoleWindowClass", NULL);    // find console handle
        if(hConsole != NULL) {
            ShowWindow(hWnd, SW_HIDE);
            ShowWindow(hConsole, SW_HIDE);
        }

    // Unix fork
    #else

      int ret;
      int i;

      ret = fork();
      if(ret != 0)
          exit(0);

      setsid();
      setpgrp();
      
      for (i=getdtablesize();i>=0;--i) 
          close(i); 
     
      i = open("/dev/null", O_RDWR); 
      dup(i); dup(i);

    #endif

    return 0;
}

//
// main()
//
int main(int argc, char *arg[])
{
    pj_status_t status;
    
    // Initialize required configuration, fetch paths
    init_config(arg[0]);
    
    // Parameter parser
    if(param_parser(argc, arg) == -1)
        return -1;
    
    // Daemonize application?
    if(oGLB.daemonized)
        daemonize_app();

    // Set log-level to 0
    pj_log_set_level(0);
  
    // Initialize pipes..    
    pipe_init();

    // Initialize signal handler
    if(signal_handler_install(argc, arg) == -1)
        return -1;

    if(!oGLB.daemonized)
    {
        printf("-------------------------------------------------------------------------------\n");
        printf(" %s %s\n", APPNAME, APPVER);
        printf("-------------------------------------------------------------------------------\n");
    }

    // Initialize pjsip
    status = pj_init();
    if(status != PJ_SUCCESS)
        goto exit;
        
    // Read in configuration
    if(read_config() == -1)
    {
        pj_log_set_level(5);
        app_err(THIS_FILE, "Cannot find dtmfbox.cfg. Please make sure the file exist!");
        goto exit;        
    }

    // set environment variables 
    {
        char env[512];        

        // DTMFBOX_CFG=dtmfbox.cfg        
        #if DTMFBOX_WIN32
        snprintf(env, sizeof(env), "DTMFBOX_CFG=%s", oGLB.configfile);
        putenv(env);
        #else
        unsetenv("DTMFBOX_CFG");
        setenv("DTMFBOX_CFG", oGLB.configfile, 1);                                        
        #endif

        // DTMFBOX_EXE=dtmfbox.exe
        #if DTMFBOX_WIN32
        snprintf(env, sizeof(env), "DTMFBOX_EXE=%s", arg[0]);
        putenv(env);
        #else
        unsetenv("DTMFBOX_EXE");
        setenv("DTMFBOX_EXE", arg[0], 1);
        #endif
    }

    // Open logfile
    if(strcmp(oGLB.logfile, ""))
    {
        oGLB.file_log = fopen(oGLB.logfile, "w+");
        if(oGLB.file_log != NULL)
            fwrite("\n\n", 2, 1, oGLB.file_log);
    }

    // Set log-level from settings and init log-callback
    {
        int log_decor = PJ_LOG_HAS_TIME | PJ_LOG_HAS_NEWLINE;
        if(oSET.logging.show_sender) log_decor |= PJ_LOG_HAS_SENDER;

        pj_log_set_level(oSET.logging.loglevel);
        pj_log_set_decor(log_decor);
        pj_log_set_log_func(&pj_log_func_cb);
    }

    // Show dtmfbox & pjsip version
    app_msg(THIS_FILE, "%s %s", APPNAME, APPVER);
    app_msg(THIS_FILE, "PJSIP %s", PJ_VERSION);

    // Create caching pool
    if(oSET.voip_active)
    {
        #if DTMFBOX_USE_VOIP

            // Initialize pjlib-util
            status = pjlib_util_init();            
            if(status != PJ_SUCCESS)
                goto exit;

            // Init PJNATH 
            status = pjnath_init();
            if(status != PJ_SUCCESS)
                goto exit;

        #endif
    }

    // Init memory pool
    pj_caching_pool_init(&oGLB.cp, &pj_pool_factory_default_policy, 0);
    oGLB.pool = pj_pool_create(&oGLB.cp.factory,        // pool factory
                           APPNAME,                     // pool name
                           POOL_SIZE_GLOBAL,            // init size
                           POOL_INCR_GLOBAL,            // increment size
                           NULL                         // callback on error
                           );

  
    // Create global endpoint
    {
        const pj_str_t *hostname;
        const char *endpt_name;

        // Global endpoint needs a unique name.
        // So we just use the hostname.
        hostname = pj_gethostname();
        endpt_name = hostname->ptr;

        // create endpoint
        status = pjsip_endpt_create(&oGLB.cp.factory, endpt_name, &oGLB.endpt);
        if(status != PJ_SUCCESS)
        {
            app_err(THIS_FILE, "(%d) Unable to create global endpoint!", status);
            goto exit;
        }
    }
   
    // Create global & account mutex 
    status = pj_mutex_create_recursive(oGLB.pool, APPNAME, &oGLB.mutex);
    if (status != PJ_SUCCESS) {
	    app_err(THIS_FILE, "Unable to create global mutex", status);
	    goto exit;
    }
    status = pj_mutex_create_recursive(oGLB.pool, NULL, &oGLB.acc_mutex);
    if (status != PJ_SUCCESS) {
	    app_err(THIS_FILE, "Unable to create account mutex", status);
	    goto exit;
    }
    
    // Init VoIP (pjsip)
    if(voip_init() != 0)
    {
        app_err(THIS_FILE, "Unable to initialize VoIP!");
        goto exit;
    }

    // Init CAPI!
    capi_init();
   
    // Initialize plugins    
    plugins_init();

    // Initialize special signal handler thread
    status = pj_thread_create(oGLB.pool, "SigHandler_%p", (pj_thread_proc*)&signal_handler_thread, (void*)NULL, 0, (unsigned)NULL, &oGLB.signal_handler_thread);
    if(status != PJ_SUCCESS)
    {
        app_err(THIS_FILE, "(%d) Unable to initialize signal handler thread!", status);
        goto exit;
    }

    // Register at all SIP-Provider(s)
    voip_client_register_init(-1);

    // Start CAPI listen loop and wait until end...
    capi_listen();
    
// ----------------------------------------------------------------------------------------------------------------------------------
//  Exit:
// ----------------------------------------------------------------------------------------------------------------------------------
exit:
    app_msg(THIS_FILE, "Exiting ...");

    // remove pidfile
    fflush(stdout);
    remove(oSET.pid_file);

    // Close logfile
    if(oGLB.file_log != NULL)
        fclose(oGLB.file_log);
    
    // Deinit plugins    
    plugins_close();

    // Deinit CAPI
    capi_close();
    
    // Deinit VoIP
    voip_close();
    app_msg(THIS_FILE, "dtmfbox exited!");
    
    // Shutdown pjsip
    pj_shutdown();
                
    // Kill process group (hanging scripts)
    #if (DTMFBOX_WIN32 == 0)    
    {
        killpg(0, 9);    
    }
    #else
    {        
        DWORD pid = GetCurrentProcessId();
        HANDLE hProc = OpenProcess(PROCESS_ALL_ACCESS, FALSE, pid);
        TerminateProcess(hProc, 9);
    }
    #endif

    exit(0);
}

//
// Initialize signal handler, check instance (parent/child)
//
int signal_handler_install(int argc, char *arg[])
{         
    int pid = getpid();
    FILE *fp;    
    
    #if DTMFBOX_WIN32
        
        char console_title[512];          

        // Create mutex
        hMutex = CreateMutex(NULL, TRUE, APPNAME);

        switch(GetLastError())
        {
            case ERROR_SUCCESS:
                // First instance
                snprintf(console_title, sizeof(console_title), "%s %s - (%d)", APPNAME, APPVER, pid);
                SetConsoleTitle(console_title);
                hWnd = FindWindow(NULL, console_title);
                break;

            case ERROR_ALREADY_EXISTS:
                // Second instance
                // read pid & get handle of already running app
                fp = fopen(oSET.pid_file, "r+");
                if(fp != NULL)
                {
                    fscanf(fp, "%d", &pid);
                    fclose(fp);

                    snprintf(console_title, sizeof(console_title), "%s %s - (%d)", APPNAME, APPVER, pid);
                    hWnd = FindWindow(NULL, console_title);
                }

                // parse/show parameters for daemon-mode
                if(param_parser_daemon(argc, arg, pid) == -1)
                    exit(1);                    

                hMutex = NULL;
                exit(0);
                break;

            default:
                break;
        }

    #endif

#if (DTMFBOX_WIN32 == 0)
    // already started?
    fp = fopen(oSET.pid_file, "r");
    if(fp != NULL)
    {
        fscanf(fp, "%d", &pid);

//#if DTMFBOX_WIN32
//        if(pid != getpid())
//#else        
        // Check if process is running under the current user.
        // When errno is EPERM, then process is running under another user
        errno = 0;        
        if(!(kill(pid, 0) == -1 && errno == ESRCH) 
            #if (DTMFBOX_USER_DEPENDANT == 0) 
                || (kill(pid, 0) == -1 && errno == EPERM) 
            #endif 
        )
//#endif
        {            
            // Already started! Show daemon-menu (actions)
            if(param_parser_daemon(argc, arg, pid) == -1)                
                exit(1);

            fclose(fp);
            return -1;
        }
        else
        {
            // Pidfile found but not started, yet.
            // Write new pid...
            fclose(fp);
            pid = getpid();

            fp = fopen(oSET.pid_file, "r+");
            if (fp == NULL) {
              app_err(THIS_FILE, "Cannot write pid-file (%s)!\n", oSET.pid_file);
              return -1;
            }
            
            rewind(fp);
            fprintf(fp, "%d\n", pid);
            fclose(fp);
            
            boost_priority(1);
        }
    }
    // Create PID file if this is the first instance
    else
    #endif
    {   
        fp = fopen(oSET.pid_file, "w+");
        if (fp == NULL) {
          app_err(THIS_FILE, "Cannot write pid-file (%s)!\n", oSET.pid_file);
          return -1;
        }
        fprintf(fp, "%d\n", getpid());     // record pid to lockfile
        fclose(fp);
        
        boost_priority(1);
    }

#if (DTMFBOX_WIN32 == 0)
    if (chmod(oSET.pid_file, 0777) != 0)    
      app_err(THIS_FILE, "Cannot change file attributes for pid-file");          
#endif

    // set the signal handler 
    signal(SIGINT,  signal_handler);        // CTRL+C
    signal(SIGTERM, signal_handler);        // Terminate
    signal(SIGHUP,  SIG_IGN);               // Ignore SIGHUP

    return 0;
}

//
// signal_handler_timer_cb()
//
// It sends an ACK msg to the SIP-Provider to keep the connection up.
// It also checks the Re-Registration of the clients/providers. If there
// is no registration within 1.5 h, the client will be removed.
// If registration to provider fails, it will retry.
//
void signal_handler_timer_cb(pj_timer_heap_t *ht, pj_timer_entry *e)
{
    #if DTMFBOX_USE_VOIP
    char c_target[MAX_URI_LENGTH], c_from[MAX_URI_LENGTH];
    pj_str_t target, from, to, contact, call_id, body;
    pj_status_t status;
    int j;
    #endif

    int i;
    int id = (int)e->user_data;
    
    PJ_UNUSED_ARG(ht);
        
    switch(id)
    {

    #if DTMFBOX_USE_VOIP

    // -----------------------------------------------------------------------------
    // Send keep-alive message to sip registrar. This will keep the connection up.
    // Simply sends an ACK
    // -----------------------------------------------------------------------------
    case 0:
        for(i=0; i<MAX_ACCOUNTS; i++)
        {
            if(oSET.accounts[i].active == 1 &&
               oSET.accounts[i].status != 0 &&
               oSET.accounts[i].controller.device != CAPI &&               
               strlen(oSET.accounts[i].registrar) &&
               strlen(oSET.accounts[i].id))
            {                
                if(oSET.voip.keep_alive != 0)
                {   
                    pjsip_tx_data *tdata;
                    pjsip_method method;
                    
                    //DTMFBOX_LOCK();

                    // Only send one ACK
                    pjsip_method_set(&method, PJSIP_ACK_METHOD);

                    pj_ansi_snprintf(c_target, sizeof(c_target), "sip:%s", oSET.accounts[i].registrar);
                    pj_ansi_snprintf(c_from, sizeof(c_target), "sip:%s", oSET.accounts[i].id);

                    target = pj_str(c_target);
                    to = pj_str(c_target);
                    from = pj_str(c_from);
                    contact = pj_str(c_from);
                    call_id = pj_str(c_from);
                    body = pj_str("!");

                    // Unbind DNS resolver temporary                    
                    #if PJSIP_HAS_RESOLVER
                        if(oGLB.dns_resolver != NULL)
                        {
                            status = pjsip_endpt_set_resolver(oGLB.endpt, NULL);
                            if (status != PJ_SUCCESS) {
                                //DTMFBOX_UNLOCK();
                                app_err(THIS_FILE, "(%d) Error unsetting DNS resolver", status);
                                goto hell;
                            }    
                        }
                    #endif
                    
                    // Create request and send keep-alive message (statelessly)
                    status = pjsip_endpt_create_request(oGLB.endpt, &method, &target, &from, &to, &contact, &call_id, 100, &body, &tdata );                
                    if(status == PJ_SUCCESS)
                        status = pjsip_endpt_send_request_stateless(oGLB.endpt, tdata, NULL, NULL);

                    if(status != PJ_SUCCESS)
                        app_err(THIS_FILE, "Sending the keep-alive has failed (%s)!", oSET.accounts[i].registrar);
                    else
                        app_dbg(THIS_FILE, "Successfully sent keep-alive (%s)", oSET.accounts[i].registrar);
                    
                    // Rebind DNS resolver temporary
                    #if PJSIP_HAS_RESOLVER
                        if(oGLB.dns_resolver != NULL)
                        {
                            status = pjsip_endpt_set_resolver(oGLB.endpt, oGLB.dns_resolver);
                            if (status != PJ_SUCCESS) {
                                //DTMFBOX_UNLOCK();
                                app_err(THIS_FILE, "(%d) Error setting DNS resolver", status);
                                goto hell;
                            }    
                        }                
                    #endif
                    
                    //DTMFBOX_UNLOCK();
                }

                // -----------------------------------------------------------------------------
                // Re-Register at Registrar, when registration failed.
                // -----------------------------------------------------------------------------
                if(oSET.accounts[i].status == 2 && oSET.accounts[i].status_nextreg == -1)
                {
                    app_msg(THIS_FILE, "Retry no. #%d to register %s", oSET.accounts[i].status_failcnt, oSET.accounts[i].registrar);
                    
                    voip_client_register_deinit(i);                    
                    pj_thread_sleep(2500);
                    voip_client_register_init(i);
                }
                else
                {
                    // reduce nextreg value (sec)
                    oSET.accounts[i].status_nextreg -= oSET.voip.keep_alive;
                }
            }
        }

        // Schedule timer again...
        {
            pj_time_val timeout;            
            timeout.msec = 0;
            if(oSET.voip.keep_alive != 0)        
                timeout.sec = oSET.voip.keep_alive;                            

            pjsip_endpt_schedule_timer(oGLB.endpt, &oGLB.keep_alive_timer, &timeout);  
        }
        break;

    // -----------------------------------------------------------------------------
    // Check registered clients table.
    // When register_inverval has timeouted, remove the client from the list.
    // The timeout is resetted, when the client re-registers.
    // -----------------------------------------------------------------------------
    case 1:
        for(i=0; i<oSET.voip.max_clients; i++)
        {
            // When client timeout is greater then register_interval (from settings),
            // then remove client from list!
            if(oGLB.clients[i].timeout >= (oSET.voip.register_interval / 60))
            {                              
                // walkthrough client accounts and reset
                for(j=0; j<MAX_ACCOUNTS; j++)
                    oGLB.clients[i].account_id[j] = -1;

                //oGLB.clients[i].addr = pj_str("");            
                pj_memset(oGLB.clients[i].addr, 0, sizeof(oGLB.clients[i].addr));
                oGLB.clients[i].timeout = 0;
            }
            else
            {                
                // increment client timeout (+1 min)
                //if(pj_strcmp2(&oGLB.clients[i].addr, ""))                                    
                if(strcmp(oGLB.clients[i].addr, ""))                                    
                    oGLB.clients[i].timeout += 1;                
            }
        }

        // Schedule timer again...
        {
            pj_time_val timeout;            
            timeout.msec = 0;            
            timeout.sec = 60;

            pjsip_endpt_schedule_timer(oGLB.endpt, &oGLB.registrar_refresh_timer, &timeout);  
        }
        break;

    #endif    

    // -----------------------------------------------------------------------------
    // Check, if there are command-line arguments to execute (second instance).
    //
    // Either parse them to dtmfbox itself (script_do_actions()) or to the plugins 
    // (plugin_raise_extra_commands()). 
    //
    // The shared memory variable "sh_actions" contains the arguments
    // -----------------------------------------------------------------------------
    case 2:        
        if(sh_actions->queue.ready == 1 && sh_actions->queue.in_use == 1)
        {   
            // default dtmfbox-commands
            if(!sh_actions->parse_plugin)
            {
                // execute actions
                script_do_actions();
           
                // reset actions
                reset_actions();
            }
            
            // plugin commands
            else  
            {
                char *args[256];
                for(i=0; i<256; i++) args[i] = sh_actions->args[i];

                // execute plugin-actions
                sh_actions->parse_plugin = plugin_raise_extra_commands(sh_actions->argc, args, sh_actions->queue.cur_pid);
            }

            // next action...
            sh_actions->queue.ready = 0;
        } 
        
        // Re-Schedule timer again...
        {
            pj_time_val timeout;            
            timeout.msec = 1;
            timeout.sec = 0;
            pjsip_endpt_schedule_timer(oGLB.endpt, &oGLB.script_event_timer, &timeout);  
        }

        break;
    }  

#if DTMFBOX_USE_VOIP
hell:
#endif
    return;
}

//
// signal_handler_thread()
//
void signal_handler_thread()
{    
    int wait_msec = 1;
    int i, j;
    pj_time_val timeout;

    // Init timer for the VoIP keep_alive callback
    #if DTMFBOX_USE_VOIP
        
        // Init timer entry
        pj_timer_entry_init(&oGLB.keep_alive_timer, 0, (void*)0, &signal_handler_timer_cb);
        pj_timer_entry_init(&oGLB.registrar_refresh_timer, 0, (void*)1, &signal_handler_timer_cb);

        // Schedule Keep-Alive timer
        if(oSET.voip.keep_alive != 0)
        {
            timeout.sec = oSET.voip.keep_alive;
            timeout.msec = 0;

            // Keep-Alive Timer
            pjsip_endpt_schedule_timer(oGLB.endpt, &oGLB.keep_alive_timer, &timeout);  
        }

        // Schedule timer to check clients for timeout (Registrar-Mode)
        timeout.sec = 60;
        pjsip_endpt_schedule_timer(oGLB.endpt, &oGLB.registrar_refresh_timer, &timeout); 

    #endif
   
    // Schedule script event timer
    {
        timeout.sec = 0;
        timeout.msec = 100;
        pj_timer_entry_init(&oGLB.script_event_timer, 0, (void*)2, &signal_handler_timer_cb);
        pjsip_endpt_schedule_timer(oGLB.endpt, &oGLB.script_event_timer, &timeout);  
    }

    // create shared memory variable (sh_actions)
    if(create_shared_memory())
    {
        app_err(THIS_FILE, "Shared memory not available!");
        exit(1);
    } else {
        sh_actions = sh_actions_shmem;
    }

    // Reset all actions
    reset_actions();
    for(i=0; i<oSET.max_connections; i++) {
        for(j=0; j<INCONF_PORTS; j++) {
            sh_actions->play_rec.play_thread[i][j] = -1;                    
        }
    }
    
    // queue counter for actions
    sh_actions->queue.in_use = 0;
    sh_actions->queue.ready = 0;
    sh_actions->queue.in_queue = 0;
    sh_actions->queue.max_queue = 0;
    sh_actions->queue.cur_queue = 0;

    // reset script event array
    pj_memset(script_events, 0, sizeof(script_events));

    // -------------------------------------------
    // SIGNAL-LOOP START:
    // -------------------------------------------
    oGLB.closing = 0;
    while(!oGLB.closing)
    {
        // Execute all pending events to script and plugins
        script_event_run();
        
        // Copy connection infos to shared memory
        //DTMFBOX_LOCK();            
        for(i=0; i<oSET.max_connections; i++)                
           pj_memcpy(&sh_actions->con_info[i], oCONS[i].info, sizeof(connection_info_t));
        
        // Check, if there are any playing ports, which got an EOF-Flag. Close these ports!            
        port_filestream_close_all_eof();                              
        //DTMFBOX_UNLOCK();
        
        // Dynamically adjust priority?
        if(oSET.adjust_priority)
            adjust_priority();

        // idle        
        pj_thread_sleep(wait_msec);
    }

    // -------------------------------------------
    // SIGNAL-LOOP END:
    // -------------------------------------------
    close_shared_memory(1);
}

//
// Receives CTRL+C and quits application (SIGTERM)
//
void signal_handler(int sig_num)
{
    // Readin ACTIONS (play file, record file, etc..)
    if(sig_num == SIGHUP)
    {
        return;
    }    
    // Exit-Application
    else
    {
        // Just run once
        if(oGLB.closing == 1)                            
            return;       

        // set closing to 1 (exit capi_listen() loop/thread)            
        oGLB.closing = 1;

        #if DTMFBOX_WIN32
            #if DTMFBOX_USE_CAPI   
                // send a listen_req, so that wait_for_signal doesn't block    
                if(oCAPI.AppId != (unsigned short)-1)            
                    LISTEN_REQ(oCAPI.cMsg, oCAPI.AppId, 0, 1, 0xff, 0x1FFF03FF, 0, NULL, NULL);                    
            #endif
        #endif
    }
}

