/*********************************************************************************
 * mixer.c ($Revision: 250 $)
 *
 * Copyright (c) 2006-2009 Marco Zissen <maz@v3v.de>
 *
 * >>> Many thanks to Florian Eyben for the mixer code! <<<
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *********************************************************************************/
#include "global_headers.h"
#include "global_objects.h"

#define THIS_FILE "mixer.c"

long vol_scale(int vol)
{
  long res;
  res = ((long)vol*(1<<F_BIAS))/100;
  if (res > 32767) res = 32767;
  if (res < 0) res = 0;
  return res;
}

int mixer_check_volume(int vol)
{
  if (vol < 0) vol = 0;
  if (vol > 32767) vol = 32767;
  return vol;
}

void mixer_convert_ms_to_frames(mixer_port_data *mixdata, long *msfr)
{  
  *msfr = (int) ( (long)mixdata->mother_port->info.clock_rate * (long)(*msfr) / ( 1000 * (long)mixdata->mother_port->info.samples_per_frame ) );
  *msfr += 1; // ensure minimum length of 1 frame
}

int mixer_init_fade(mixer_port_data *mixdata, int target_vol, int fadetime, int is_rx) 
{  
  if(is_rx)
  {
    // set target rx volume
    mixdata->target_rx_volume = mixer_check_volume(target_vol);
    mixdata->fade_rx_frames_remain = fadetime;
    mixer_convert_ms_to_frames(mixdata, &mixdata->fade_rx_frames_remain);
  }
  else
  {
    // set target tx volume
    mixdata->target_tx_volume = mixer_check_volume(target_vol);
    mixdata->fade_tx_frames_remain = fadetime;
    mixer_convert_ms_to_frames(mixdata, &mixdata->fade_tx_frames_remain);
  }

  mixdata->infade = 1;

  return 1; 
}

/** internal functions **/

__inline pj_int16_t abs16(pj_int16_t x)
{
  if (x<0) return -x;
  else return x;
}

void recompute_talkover_active(int con_id)
{
  int tmp = 0;
  int i;

  for (i=0; i<N_CON_PORTS; i++) {
    if ( (oCONS[con_id].mixer->talkover_active[i]) 
         && (oCONS[con_id].mixer->is_talkover[i])
       )
    {
      tmp = 1;
      break;
    }
  }

  oCONS[con_id].mixer->in_talkover = tmp;
}

void set_talkover_active(int con_id, int play_id, int value)
{
  oCONS[con_id].mixer->talkover_active[play_id] = value;
  recompute_talkover_active(con_id);
}

// input: log. volume 0..256
// output: lin. voluem 0..256 ready to be applied to samples

// log: can be applied to samples, lin: for the user
// range lin: 0..128..256  (=0-100%-200%)
// range log: 0..128..512
__inline long vol_lin2log(long vol) 
{
  return vol;
  // return (long)( powf(((float)vol)/(float)(1<<F_BIAS), VOL_EXPONENT) * 128.0 );
  // return (int)(log((float)(vol+1))*(128.0/log(128.0)) );
}

__inline long vol_log2lin(long vol) 
{
  return vol;
  // return (int)( powf(((float)vol)/(float)(1<<F_BIAS),1.0/VOL_EXPONENT) * (float)(1<<F_BIAS) )+1;
  // return (int)( exp(vol*log(128.0)/128.0)-1.0 )+1;  
}


// factor should be between 0..32767 , sample is 16-bit
__inline void mixer_vol_mult(pj_int16_t *sample, long factor)
{   
    *sample = (pj_int16_t)(((long)(*sample) * (factor) ) >> F_BIAS);
    //*sample = ( (long)(*sample) * (factor) ) >> F_BIAS;
}
// TODO: loudness compressor

void do_mixing(mixer_port_data *mixdata, pjmedia_frame *frame, int con_id, int play_id, int rec_id)
{
    pj_int16_t *samples = frame->buf;
    int id = -1;
    int size = 0;
    int is_rx = 0;
    int target_volume;
    int volume;
    int fade_remain;

    // play or rec id?
    if(play_id != -1) id = play_id;    
    if(rec_id != -1) id = rec_id;
    
    // RX or TX?
    is_rx = (play_id != -1);
    if(is_rx)
    {
        volume = mixdata->rx_volume;
        target_volume = mixdata->target_rx_volume;        
        fade_remain = mixdata->fade_rx_frames_remain;
    }
    else
    {
        volume = mixdata->tx_volume;
        target_volume = mixdata->target_tx_volume;        
        fade_remain = mixdata->fade_tx_frames_remain;
    }

    size = frame->size / (sizeof(pj_int16_t));

    //---- do mixing !-----
    if (oCONS[con_id].mixer->is_talkover[id]) {
       // compute frame's simplified "rms" volume (simple mean)
       long mean = 0; 
       pj_int16_t *p1, *p2;
       p1 = p2 = samples; p2 += size;
       while (p1<p2) {
           mean += abs16(*p1); 
           p1++;
       }
       mean /= size;

       if (mean > SILENCE_THRESHOLD) {
           oCONS[con_id].mixer->talkover_active[id] = 1;
       } else {
           oCONS[con_id].mixer->talkover_active[id] = 0;
       }
    }

    recompute_talkover_active(con_id);

    if ( (oCONS[con_id].mixer->is_background[id])
        &&(oCONS[con_id].mixer->in_talkover==1)
        &&(!mixdata->talkover_fade_active) ) {
       // begin of talkover -> reduce volume: 
       if (mixdata->infade) {
          mixdata->notalkover_rx_volume = target_volume;              
       } else {
          mixdata->notalkover_rx_volume = volume;
       }

       mixdata->talkover_fade_active = 1;
       mixdata->talkover_downfade = 1;
       mixer_init_fade(mixdata, vol_scale(TALKOVER_VOLUME), 50, is_rx); // last parameter 0: don't lock mixer
    }

    if ( (mixdata->talkover_fade_active)
        && ( (oCONS[con_id].mixer->in_talkover==0)
        || (!oCONS[con_id].mixer->is_background[id]) )
      ) {
        mixdata->talkover_fade_active = 0;
        mixer_init_fade(mixdata, mixdata->notalkover_rx_volume, 300, is_rx); // last parameter 0: don't lock mixer
        mixdata->notalkover_rx_volume = 0;
    }

    if (mixdata->infade && fade_remain) 
    {           
       // fade linear per frame, compute start & end values
       // 32 bit integer arithmetics on 16 bit sample values 
       // start vol == rx_volume
       if (fade_remain) {
           long start_vol = (long)(volume);
           long rx_lin = vol_log2lin(volume);
           long end_vol = vol_lin2log( (long)(rx_lin) + (((long)(target_volume) - (long)(rx_lin))) / (long)(fade_remain) );
           long fade_step = (end_vol - start_vol); // /size;

           if ( (!(mixdata->talkover_fade_active))
               || (!oCONS[con_id].mixer->is_background[id]) 
               || (mixdata->talkover_downfade) ) {
            // the actual fade
               long cur_vol = start_vol*size;

               pj_int16_t *p1, *p2;
               p1 = p2 = samples; p2 += size;
               while(p1<p2) {
                   mixer_vol_mult(p1, cur_vol/size);
    //                   *p1 = ( ((long)(*p1) * (cur_vol/size)) >> F_BIAS);
                   p1++;
                   cur_vol += fade_step;
               }

               if(is_rx)
                  mixdata->rx_volume = (end_vol);
               else
                  mixdata->tx_volume = (end_vol);
           } 

           else // fade notalkover_rx_volume in background, talkover currently active
           {
               pj_int16_t *p1, *p2;

               mixdata->notalkover_rx_volume = end_vol;
               p1 = p2 = samples; p2 += size;
               while(p1<p2) {
                   mixer_vol_mult(p1, volume);
    //                     *p1 = ( (long)*p1 * (mixdata->rx_volume) ) >> F_BIAS;
                   p1++;
               }
           }

           // RX
           if(is_rx)
           {
               mixdata->fade_rx_frames_remain--; 
               if (mixdata->fade_rx_frames_remain <= 0) {
                   mixdata->infade = 0;
                   mixdata->fade_rx_frames_remain = 0;
                   if (mixdata->talkover_downfade)
                       mixdata->talkover_downfade = 0;
               }
           }
           // TX
           else
           {
               mixdata->fade_tx_frames_remain--; 
               if (mixdata->fade_tx_frames_remain <= 0) {
                   mixdata->infade = 0;
                   mixdata->fade_tx_frames_remain = 0;
                   if (mixdata->talkover_downfade)
                       mixdata->talkover_downfade = 0;
               }
           }

       } else { 
           mixdata->infade = 0; // this case should never happen!
       }
    }           

    else   // not in fade
    {
       int default_volume = MIX_DEF_RX_VOLUME;
       if(!is_rx) default_volume = MIX_DEF_TX_VOLUME;

       if(vol_lin2log(vol_scale(default_volume)) != volume)     // default volume? skip...
       {
           pj_int16_t *p1, *p2;
           samples = frame->buf;
           p1 = p2 = samples; p2 += size;
           
           while(p1<p2) {
               mixer_vol_mult(p1, volume);
               //*p1 = ( (long)*p1 * (mixdata->rx_volume) ) >> F_BIAS;
               p1++;
           }
       }           
    }
}

/**** public API *****/


// Every getframe()-callback calls this function before writing to downstream port.
pj_status_t mixer_port_get_frame(pjmedia_port *port, pjmedia_frame *frame, int con_id, int play_id, int dn_status)
{    
    pj_status_t status = PJ_SUCCESS;
    mixer_port_data *mixdata = NULL;
    //int i;

     //return status;  // for debugging: disable mixing!

    // port_capi.c, port_voip.c, port_soundcard.c
    if((!pj_strcmp2(&port->info.name, "capi port") || 
        !pj_strcmp2(&port->info.name, "voip port") ||
        !pj_strcmp2(&port->info.name, "soundcard port")))
    {
        portdata_capi *pdata_c = NULL; 
        portdata_voip *pdata_v = NULL;
        portdata_soundcard *pdata_s = NULL;

        switch(oCONS[con_id].info->type)
        {
            case CAPI:   
                pdata_c = (portdata_capi *)(port->port_data.pdata);
                if (pdata_c) mixdata = (mixer_port_data *) ( &(pdata_c->mixdata) );
                play_id = PLID_CON_RX + con_id;
                break;

            case VOIP:
                pdata_v = (portdata_voip *)(port->port_data.pdata);
                if (pdata_v) mixdata = (mixer_port_data *) ( &(pdata_v->mixdata) );
                play_id = PLID_CON_RX + con_id;
                break;

            case SCRD:
                pdata_s = (portdata_soundcard *)(oCONS[con_id].ports->custom->port_data.pdata);
                //pdata_s = (portdata_soundcard *)(port->port_data.pdata);
                if (pdata_s) mixdata = (mixer_port_data *) ( &(pdata_s->mixdata) );
                play_id = PLID_CON_RX + con_id;
                break;
        }
    }

    // port_filestream.c             
    else if(play_id != -1)
    {   
        portdata_filestream *pdata_f = (portdata_filestream *)(port->port_data.pdata);
        if (pdata_f) mixdata = (mixer_port_data *) ( &(pdata_f->mixdata) );
    }

//    if((play_id > -1)&&(play_id<INCONF_PORTS))
    if (play_id > -1)
    {
        if ((dn_status != PJ_SUCCESS)&&(oCONS[con_id].mixer->is_talkover[play_id])) {
            set_talkover_active(con_id, play_id, 0);
        }

        if (mixdata) 
            do_mixing(mixdata, frame, con_id, play_id, -1);
    }

    // raise plugin event "plugin_get_frame()"
    plugin_raise_get_frame(frame->buf, frame->size, con_id, (play_id >= PLID_CON_RX ? -1 : play_id), oCONS[con_id].info->account_id);

    return status;
}

pj_status_t mixer_port_put_frame(pjmedia_port *port, pjmedia_frame *frame, int con_id, int rec_id)
{    
    pj_status_t status = PJ_SUCCESS;
    mixer_port_data *mixdata = NULL;

    //return status;  // for debugging: disable mixing!

    // port_capi.c, port_voip.c, port_soundcard.c
    if((!pj_strcmp2(&port->info.name, "capi port") || 
        !pj_strcmp2(&port->info.name, "voip port") ||
        !pj_strcmp2(&port->info.name, "soundcard port")) && rec_id == -1)
    {
        portdata_capi *pdata_c = NULL; 
        portdata_voip *pdata_v = NULL;
        portdata_soundcard *pdata_s = NULL;

        switch(oCONS[con_id].info->type)
        {            
            case CAPI: 
                pdata_c = (portdata_capi *)(port->port_data.pdata);
                if (pdata_c) mixdata = (mixer_port_data *) ( &(pdata_c->mixdata) );
                rec_id = PLID_CON_TX + con_id;
                break;

            case VOIP:
                pdata_v = (portdata_voip *)(port->port_data.pdata);
                if (pdata_v) mixdata = (mixer_port_data *) ( &(pdata_v->mixdata) );
                rec_id = PLID_CON_TX + con_id;
                break;

            case SCRD:
                pdata_s = (portdata_soundcard *)(oCONS[con_id].ports->custom->port_data.pdata);
                if (pdata_s) mixdata = (mixer_port_data *) ( &(pdata_s->mixdata) );
                rec_id = PLID_CON_TX + con_id;
                break;
        }
    }

    if(rec_id == 0)
    {
        portdata_filestream *pdata_f = NULL;                
        mixdata = NULL;

        if(port) 
            pdata_f = (portdata_filestream *)(port->port_data.pdata);

        if (pdata_f)
            mixdata = (mixer_port_data *) ( &(pdata_f->mixdata) );        

        if(mixdata)
            rec_id = PLID_REC_TX;        
    }

    if(mixdata != NULL)
    {         
        if (oCONS[con_id].mixer->is_talkover[rec_id]) {
            set_talkover_active(con_id, rec_id, 0);
        }
        
        if (mixdata) 
            do_mixing(mixdata, frame, con_id, -1, rec_id);
    }

    // raise plugin event "plugin_put_frame()"
    plugin_raise_put_frame(frame->buf, frame->size, con_id, (rec_id == PLID_REC_TX ? 0 : -1), oCONS[con_id].info->account_id);

    return status;
}


// to set capi/voip/... rx/tx mixdata, use as play_id: PLID_CAPI_RX , PLID_VOIP....
void script_mixer_control(int play_id, int rec_id)
{
  char info[256];
  mixer_port_data *mixdata = NULL;
  portdata_filestream *pdata = NULL;  
  int sh_conid = sh_actions->connection;
  int new_player = 0;
  int new_recorder = 0;
  int is_rx = 0;

  // flags, to check if option was set, for applying defaults at end
  int fVol=0,fTo=0,fBg=0;

  // no "-mixer <play_id>" and no "-play... -mixer" option
  if ((sh_actions->mixer.play_id == -1) && (play_id == -1) &&
      (sh_actions->mixer.rec_id == -1) && (rec_id == -1)) {

      return;
  }
  
  if ((sh_actions->mixer.play_id == -1) && (play_id >= 0)) {
    new_player = 1;
    sh_actions->mixer.play_id = play_id;
  } 
  else
    play_id = sh_actions->mixer.play_id;

  if ((sh_actions->mixer.rec_id == -1) && (rec_id >= 0)) {
    new_recorder = 1;
    sh_actions->mixer.rec_id = rec_id;
  }
  else 
      rec_id = sh_actions->mixer.rec_id;
  
  // check if play_id is in valid range
  if ( ((sh_actions->mixer.play_id < 0) || (sh_actions->mixer.play_id > (N_CON_PORTS))) && 
       ((sh_actions->mixer.rec_id < 0) || (sh_actions->mixer.rec_id > (N_CON_PORTS)))) 
  {
    char info[256];
    app_dbg(THIS_FILE, "%s Mixer => INVALID play_id/rec_id (play_id=%d, rec_id=%d)!", connection_print_info(&oCONS[sh_conid], info, sizeof(info)), sh_actions->mixer.play_id, sh_actions->mixer.rec_id);
    return;
  }

  // RX or TX?
  is_rx = (play_id != -1);

  // check for 'special' play_ids / rec_ids:
  // these are connections (tx/rx)
  if ((sh_actions->mixer.play_id >= INCONF_PORTS  || sh_actions->mixer.rec_id >= INCONF_PORTS) &&
      (sh_conid >= 0 && sh_conid < INCONF_PORTS)) 
  {   
    portdata_capi *pdata_c = NULL;
    portdata_voip *pdata_v = NULL;
    portdata_soundcard *pdata_s = NULL;

    switch(oCONS[sh_conid].info->type)
    {
        // capi port
        case CAPI:    
            pdata_c = (portdata_capi *)(oCONS[sh_conid].ports->custom->port_data.pdata);
            if (pdata_c) mixdata = (mixer_port_data *) ( &(pdata_c->mixdata) );            
            break;

        // voip port
        case VOIP:
            pdata_v = (portdata_voip *)(oCONS[sh_conid].ports->custom->port_data.pdata);
            if (pdata_v) mixdata = (mixer_port_data *) ( &(pdata_v->mixdata) );            
            break;

        // soundcard port
        case SCRD:
            pdata_s = (portdata_soundcard *)(oCONS[sh_conid].ports->custom->port_data.pdata);
            if (pdata_s) mixdata = (mixer_port_data *) ( &(pdata_s->mixdata) );            
            break;            
    }
  }
  else 
  {
      int mixer_play_id = sh_actions->mixer.play_id;
      int mixer_rec_id = sh_actions->mixer.rec_id;

      // check if play port with play_id exists
      if(mixer_play_id != -1)
      {      
          if (oCONS[sh_conid].ports->play_stream[mixer_play_id]) {
            pdata = (portdata_filestream *)(oCONS[sh_conid].ports->play_stream[mixer_play_id]->port_data.pdata);
            mixdata = (mixer_port_data *) ( &(pdata->mixdata) );            
          }
      }    

      // check if rec_port exists
      if(mixer_rec_id != -1)
      {
          if (oCONS[sh_conid].ports->rec_stream) {
            pdata = (portdata_filestream *)(oCONS[sh_conid].ports->rec_stream->port_data.pdata);
            mixdata = (mixer_port_data *) ( &(pdata->mixdata) );
          }
      }
  }

  if (mixdata) {
   
    // set volume
    if(sh_actions->mixer.options & OPT_MIX_VOLUME) 
    {
        if(sh_actions->mixer.volume > -1) {
            fVol = 1;
            if (mixer_init_fade(mixdata, vol_lin2log(vol_scale(sh_actions->mixer.volume)), 100, is_rx)) 
                app_dbg(THIS_FILE, "%s Set volume to %i (play_id=%d, rec_id=%d)", connection_print_info(&oCONS[sh_conid], info, sizeof(info)), mixer_check_volume(sh_actions->mixer.volume), sh_actions->mixer.play_id, sh_actions->mixer.rec_id);
        }
    }

    // set fading
    if(sh_actions->mixer.options & OPT_MIX_FADE) 
    {             
        if(sh_actions->mixer.fade_volume > -1) {
            fVol = 1;
            if (mixer_init_fade(mixdata, vol_lin2log(vol_scale(sh_actions->mixer.fade_volume)), sh_actions->mixer.fade_duration, is_rx))
                app_dbg(THIS_FILE, "%s Start fade to vol %i over %i ms (play_id=%d, rec_id=%d)", connection_print_info(&oCONS[sh_conid], info, sizeof(info)), mixer_check_volume(sh_actions->mixer.fade_volume), sh_actions->mixer.fade_duration, sh_actions->mixer.play_id, sh_actions->mixer.rec_id);
        }
    }

    // set talkover on/off
    if((sh_actions->mixer.options & OPT_MIX_TALKOVER_ON) || 
       (sh_actions->mixer.options & OPT_MIX_TALKOVER_OFF)) 
        {
            int tk_on = 0;
            fTo = 1;
            if(sh_actions->mixer.options & OPT_MIX_TALKOVER_ON) {
                if(sh_actions->mixer.play_id != -1) {
                    tk_on = sh_actions->mixer.play_id;
                    oCONS[sh_conid].mixer->is_talkover[sh_actions->mixer.play_id] = 1;
                    oCONS[sh_conid].mixer->is_background[sh_actions->mixer.play_id] = 0;
                    set_talkover_active(sh_conid, sh_actions->mixer.play_id, 0);          
                }
                if(sh_actions->mixer.rec_id != -1) {
                    tk_on = sh_actions->mixer.rec_id;
                    oCONS[sh_conid].mixer->is_talkover[sh_actions->mixer.rec_id] = 1;
                    oCONS[sh_conid].mixer->is_background[sh_actions->mixer.rec_id] = 0;
                    set_talkover_active(sh_conid, sh_actions->mixer.rec_id, 0);          
                }
            }
            
            if(sh_actions->mixer.options & OPT_MIX_TALKOVER_OFF) {
                if(sh_actions->mixer.play_id != -1) {
                    tk_on = sh_actions->mixer.play_id;
                    oCONS[sh_conid].mixer->is_talkover[sh_actions->mixer.play_id] = 0;
                    set_talkover_active(sh_conid, sh_actions->mixer.play_id, 0);
                }
                if(sh_actions->mixer.rec_id != -1) {
                    tk_on = sh_actions->mixer.rec_id;
                    oCONS[sh_conid].mixer->is_talkover[sh_actions->mixer.rec_id] = 0;
                    set_talkover_active(sh_conid, sh_actions->mixer.rec_id, 0);                
                }

            }
            
            app_dbg(THIS_FILE, "%s Set talkover=%d (play_id=%d, rec_id=%d)", connection_print_info(&oCONS[sh_conid], info, sizeof(info)), oCONS[sh_conid].mixer->is_talkover[tk_on], sh_actions->mixer.play_id, sh_actions->mixer.rec_id);
        }

    // set background on/off
    if((sh_actions->mixer.options & OPT_MIX_BACKGROUND_ON) || 
       (sh_actions->mixer.options & OPT_MIX_BACKGROUND_OFF)) 
        {
            int bg_on = 0;
            fBg = 1;

            if(sh_actions->mixer.options & OPT_MIX_BACKGROUND_ON) {
                if(sh_actions->mixer.play_id != -1) {
                    bg_on = sh_actions->mixer.play_id;
                    oCONS[sh_conid].mixer->is_background[sh_actions->mixer.play_id] = 1;
                    oCONS[sh_conid].mixer->is_talkover[sh_actions->mixer.play_id] = 0;
                    set_talkover_active(sh_conid, sh_actions->mixer.play_id, 0);                    
                }
                if(sh_actions->mixer.rec_id != -1) {
                    bg_on = sh_actions->mixer.rec_id;
                    oCONS[sh_conid].mixer->is_background[sh_actions->mixer.rec_id] = 1;
                    oCONS[sh_conid].mixer->is_talkover[sh_actions->mixer.rec_id] = 0;
                    set_talkover_active(sh_conid, sh_actions->mixer.rec_id, 0);                    
                }                                
            }

            if(sh_actions->mixer.options & OPT_MIX_BACKGROUND_OFF) {
                if(sh_actions->mixer.play_id != -1) {
                    bg_on = sh_actions->mixer.play_id;
                    oCONS[sh_conid].mixer->is_background[sh_actions->mixer.play_id] = 0;                                        
                }
                if(sh_actions->mixer.rec_id != -1) {
                    bg_on = sh_actions->mixer.rec_id;
                    oCONS[sh_conid].mixer->is_background[sh_actions->mixer.rec_id] = 0;
                }                
            }

            app_dbg(THIS_FILE, "%s Set background=%d (play_id=%d, rec_id=%d)", connection_print_info(&oCONS[sh_conid], info, sizeof(info)), oCONS[sh_conid].mixer->is_background[bg_on], sh_actions->mixer.play_id, sh_actions->mixer.rec_id);
        }

    // initialize a new player with defaults:    
    if (new_player == 1) 
    {
       // fade or volume
       if (!fVol) 
           mixer_init_fade(mixdata, vol_lin2log(vol_scale(oSET.audio.mixer_rx_volume)), 100, 1);
       
       // background if play_id == 0
       if ((sh_actions->mixer.play_id == 0) && (!fBg) && (!fTo)) {
           oCONS[sh_conid].mixer->is_background[sh_actions->mixer.play_id] = 1;
           oCONS[sh_conid].mixer->is_talkover[sh_actions->mixer.play_id] = 0;
           set_talkover_active(sh_conid, sh_actions->mixer.play_id,0);          
       } 

       // talkover if play_id > 0       
       else if ( (!fTo) && (!fBg) && oSET.audio.mixer_talkover ) {
           oCONS[sh_conid].mixer->is_talkover[sh_actions->mixer.play_id] = 1;
           oCONS[sh_conid].mixer->is_background[sh_actions->mixer.play_id] = 0;
           set_talkover_active(sh_conid, sh_actions->mixer.play_id,0);          
       }
    }
  }
}
