/*********************************************************************************
 * voip_auth.c ($Revision: 261 $)
 *
 * Copyright (c) 2006-2009 Marco Zissen <maz@v3v.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *********************************************************************************/
#include "global_headers.h"
#include "global_objects.h"

#define THIS_FILE "voip_auth.c"

#if DTMFBOX_USE_VOIP

// ------------------------------------------------------------------------------------------------------------
//
//  AUTHORIZATION (CLIENT -> DTMFBOX)
//
// ------------------------------------------------------------------------------------------------------------

static void digest2str(const unsigned char digest[], char *output)
{
    int i;
    for (i = 0; i<16; ++i)
    {
        pj_val_to_hex_digit(digest[i], output);
        output += 2;
    }
}

//
// Request a authorization header (www-authorization), when not given.
// After sending this message, an authorization header will be received.
//
int voip_registrar_req_auth_header(pjsip_rx_data *rdata)
{
    pjsip_hdr hdr_list;
    pjsip_www_authenticate_hdr *hdr;

    pj_str_t reason = pj_str("need authorization challenge!");
    pj_str_t realm = pj_str((char*)oSET.voip.realm);
    char nonce_buf[16];    
    pj_str_t random;
    
    // No realm? Set current IP address!
    if(realm.slen == 0) 
    {   
        realm = pj_str(oSET.voip.ip_addr);
        realm.slen = strlen(oSET.voip.ip_addr);
        app_dbg2(THIS_FILE, "No realm in dtmfbox.cfg specified! Using %s", oSET.voip.ip_addr);
    }

    app_msg(THIS_FILE, "Request authorization challenge (realm: %s) ...", realm);
    
    random.ptr = nonce_buf;
    random.slen = sizeof(nonce_buf);
    pj_memset(nonce_buf, 0, sizeof(nonce_buf));

    // Create the auth-header & contact header
    hdr = pjsip_www_authenticate_hdr_create(rdata->tp_info.pool);

    // Initialize auth-header
    // Only digest authentication (md5)!
    hdr->scheme = pjsip_DIGEST_STR;
    hdr->challenge.digest.algorithm = pjsip_MD5_STR;
    pj_create_random_string(nonce_buf, sizeof(nonce_buf));
    pj_strdup(rdata->tp_info.pool, &hdr->challenge.digest.nonce, &random);
    pj_create_random_string(nonce_buf, sizeof(nonce_buf));
    pj_strdup(rdata->tp_info.pool, &hdr->challenge.digest.opaque, &random);
    hdr->challenge.digest.qop.slen = 0;
    pj_strdup(rdata->tp_info.pool, &hdr->challenge.digest.realm, &realm);
    hdr->challenge.digest.stale = 0;
    pj_strdup(rdata->tp_info.pool, &hdr->challenge.common.realm, &realm);

    // Push hdr to hdr_list
    pj_list_init(&hdr_list);    
    pj_list_push_back(&hdr_list, (pjsip_hdr*)oGLB.useragent_hdr);       // USER-AGENT HEADER    
    pj_list_push_back(&hdr_list, (pjsip_www_authenticate_hdr*)hdr);     // WWW-AUTH HEADER

    // Respond with statuscode 401 (not authorized) and pass hdr_list..
    // This will initialize the authorization challenge
    pjsip_endpt_respond_stateless(oGLB.endpt, rdata, 401, &reason, &hdr_list, NULL);

    return 0;
}


//
// Search for the account in app-settings and create a pjsip_cred_info object
//
pj_status_t voip_registrar_lookup_account(pj_pool_t *pool, const pj_str_t *realm, const pj_str_t *acc_name, pjsip_cred_info *cred_info)
{
    char cur_realm[256];
    char cur_user[256];
    int i;

    PJ_UNUSED_ARG(pool);

    // copy account and name from params to local variables
    pj_memcpy(&cur_realm, realm->ptr, sizeof(cur_realm));
    pj_memcpy(&cur_user, acc_name->ptr, sizeof(cur_user));

    if(acc_name->slen < 256)
        cur_user[acc_name->slen] = '\0';
    if(realm->slen < 256)
        cur_realm[realm->slen] = '\0';

    i = account_by_username(cur_user);

    if(i != -1)
    {
        app_msg(THIS_FILE, "Found credentials -> user: %s, realm: %s", cur_user, cur_realm);

        cred_info->realm     = pj_str(cur_realm);
        cred_info->username  = pj_str(cur_user);
        cred_info->data_type = PJSIP_CRED_DATA_PLAIN_PASSWD;
        cred_info->scheme =    pj_str("digest");
        cred_info->data      = pj_str(oSET.accounts[i].registrar_password);

        return PJ_SUCCESS;
    }

    app_msg(THIS_FILE, "No matching credentials found!");
    return PJSIP_EAUTHACCNOTFOUND;
}


//
// Verify incoming authorization header against the previously created credential.
// RFC 2617
//
pj_status_t voip_registrar_verify_account( const pjsip_authorization_hdr *hdr, const pj_str_t *method, const pjsip_cred_info *cred_info)
{
    const pjsip_digest_credential *dig = &hdr->credential.digest;
    char ha1[PJSIP_MD5STRLEN];
    char ha2[PJSIP_MD5STRLEN];
    unsigned char digest[16]; // PJSIP_MD5STRLEN
    unsigned char username[64];
    unsigned char password[64];
    unsigned char realm[64];    
    pj_str_t final_digest;
    pj_md5_context pms;

    pj_memcpy(&realm, cred_info->realm.ptr, sizeof(realm));
    pj_memcpy(&username, cred_info->username.ptr, sizeof(username));
    pj_memcpy(&password, cred_info->data.ptr, sizeof(password));

    if (pj_stricmp(&hdr->scheme, &pjsip_DIGEST_STR) == 0)
    {
        // username : realm : password    -> ha1
        // -----------------------------------
        if (cred_info->data_type == PJSIP_CRED_DATA_PLAIN_PASSWD)
        {
            pj_md5_init(&pms);
            MD5_APPEND(&pms, (const unsigned char*)username, strlen(username));            
            MD5_APPEND(&pms, ":", 1);

            // maz: I only get the client's realm (ip) :(
            // .. so just replaced it with header-realm ...
            // MD5_APPEND( &pms, &realm, cred_info->realm.slen);
            MD5_APPEND( &pms, hdr->credential.digest.realm.ptr, hdr->credential.digest.realm.slen);

            MD5_APPEND( &pms, ":", 1);            
            MD5_APPEND( &pms, (const unsigned char*)password, strlen(password));
            pj_md5_final(&pms, digest);
            digest2str(digest, ha1);
            //ha1[PJSIP_MD5STRLEN] = '\0';
            digest[0] = '\0';

        }
        else if (cred_info->data_type == PJSIP_CRED_DATA_DIGEST)
        {
            pj_assert(cred_info->data.slen == 32);            
            pj_memcpy( ha1, (const unsigned char*)password, strlen(password));
        }

        // method : uri        -> ha2
        // -----------------------
        pj_md5_init(&pms);
        MD5_APPEND( &pms, method->ptr, method->slen);
        MD5_APPEND( &pms, ":", 1);
        MD5_APPEND( &pms, dig->uri.ptr, dig->uri.slen);
        pj_md5_final(&pms, digest);
        digest2str(digest, ha2);
        //ha2[PJSIP_MD5STRLEN] = '\0';
        digest[0] = '\0';

        // ha1 : nonce : [ nc:cnonce:qop ] : ha2    -> final_digest
        // --------------------------------------------------------
        pj_md5_init(&pms);
        MD5_APPEND( &pms, ha1, PJSIP_MD5STRLEN);
        MD5_APPEND( &pms, ":", 1);
        MD5_APPEND( &pms, dig->nonce.ptr, dig->nonce.slen);
        MD5_APPEND( &pms, ":", 1);

        if (dig->qop.slen != 0) {

            MD5_APPEND( &pms, dig->nc.ptr, dig->nc.slen);
            MD5_APPEND( &pms, ":", 1);
            MD5_APPEND( &pms, dig->cnonce.ptr, dig->cnonce.slen);
            MD5_APPEND( &pms, ":", 1);
            MD5_APPEND( &pms, dig->qop.ptr, dig->qop.slen);
            MD5_APPEND( &pms, ":", 1);
        }

        MD5_APPEND( &pms, ha2, PJSIP_MD5STRLEN);
        pj_md5_final(&pms, digest);
        digest2str(digest, ha1);
        //ha1[PJSIP_MD5STRLEN] = '\0';

        final_digest = pj_str(ha1);
        final_digest.slen = PJSIP_MD5STRLEN;

        app_dbg2(THIS_FILE, "DIGEST=HEADER %.*s=%.*s", final_digest.slen, final_digest.ptr, hdr->credential.digest.response.slen, hdr->credential.digest.response.ptr);
            
        // Compare digest.
        return (pj_stricmp(&final_digest, &hdr->credential.digest.response) == 0) ? PJ_SUCCESS : PJSIP_EAUTHINVALIDDIGEST;
    }
    else
    {
        return PJSIP_EINVALIDAUTHSCHEME;
    }
}

//
// Main function, which checks the received message for authorization header.
//
// When no authorization header is found, the app sends a 401 respond. After that, the
// client should send authentication header. This header is MD5 encrypted. So we
// just encrypt our credentials and compare it with the received. That's it!
//
// This function returns the status-code (200 Ok / 401 Not Authorized / 500 Unknown)
//
int voip_registrar_check_auth_header(pjsip_rx_data* rdata, int *acc_id)
{
    pj_str_t realm = pj_str(oSET.voip.realm);
    pjsip_cred_info cred_info;
    pjsip_authorization_hdr* hdr;
    pjsip_msg *msg = rdata->msg_info.msg;
    pj_str_t acc_name;
    pj_status_t status;
    int statuscode = 200;

    *acc_id = -1;

    // Get WWW-Authorization part from header
    hdr = (pjsip_authorization_hdr*) pjsip_msg_find_hdr(msg, PJSIP_H_AUTHORIZATION, NULL);

    // No realm? Set current IP address!
    if(realm.slen == 0) 
    {   
        realm = pj_str(oSET.voip.ip_addr);
        realm.slen = strlen(oSET.voip.ip_addr);
        app_dbg2(THIS_FILE, "No realm in dtmfbox.cfg specified! Using %s", oSET.voip.ip_addr);
    }
    
    if(hdr)
    {
        // Get realm from client
        while (hdr) 
        {   
            // digest realm?
            if(hdr->credential.digest.realm.slen > 0) {
                pj_strcpy(&realm, &hdr->credential.digest.realm);
                break;
            }

            // common realm?
            if(hdr->credential.common.realm.slen > 0) {
                pj_strcpy(&realm, &hdr->credential.common.realm);
                break;
            }

            hdr = hdr->next;
            if (hdr == (void*) &msg->hdr) {
                hdr = NULL;
                break;
            }

            hdr = (pjsip_authorization_hdr*)pjsip_msg_find_hdr(msg, PJSIP_H_AUTHORIZATION, hdr);
        }

        // Got header to realm?
        if(!hdr)
        {
            app_err(THIS_FILE, "Unable to find realm!");
            return 401;
        }

        // Check authorization scheme
        if (pj_stricmp(&hdr->scheme, &pjsip_DIGEST_STR) == 0)
        {
            acc_name = hdr->credential.digest.username;
        }
        else
        {
            app_err(THIS_FILE, "Wrong authentication scheme!");
            return 401;
        }

        // Lookup credential in oSET
        status = voip_registrar_lookup_account(rdata->tp_info.pool, &realm, &hdr->credential.digest.username, &cred_info);
        if (status != PJ_SUCCESS)
        {
            app_err(THIS_FILE, "User not found!");
            return 401;
        }

        // Authenticate with the specified credential.
        status = voip_registrar_verify_account(hdr, &msg->line.req.method.name, &cred_info);
        if (status != PJ_SUCCESS)
        {
            app_err(THIS_FILE, "Wrong username/password!");
            return 401;
        }
    }
    else
    {
        statuscode = 500;
    }

    // Put successful registrations to oGLB.clients array
    // Later we can decide, if an INVITE session comes from the registrar or the client...
    //
    if(statuscode==200)
    {        
        char client[MAX_URI_LENGTH];
        char username[MAX_URI_LENGTH];
        pjsip_contact_hdr *client_contact;
        pj_str_t cur_client;
        int add = 0;
        int i, j;
        int macc_id = -1;

        // save "server:port" and username
        pj_ansi_snprintf(client, sizeof(client), "%s:%d", rdata->pkt_info.src_name, rdata->pkt_info.src_port);        
        pj_ansi_snprintf(username, sizeof(username), "%.*s", hdr->credential.digest.username.slen, hdr->credential.digest.username);
        cur_client = pj_str(client);        
        
        // get account by username
        macc_id = account_by_username(username);

        // get client-id by address...
        i = client_by_address(client);

        // ..not found? get free id
        if(i == -1)
            i = client_by_free();

        // .. no free client left :(
        if(i == -1)
        {
            app_err(THIS_FILE, "Unable to authorize client %s! Max. %d client(s) allowed!", client, oSET.voip.max_clients);
            statuscode = 487;
            return statuscode;
        }

        // Get client account by account
        j = client_account_by_account(i, macc_id);

        // account already added to client
        if(j != -1)
        {
            add = 1;
            oGLB.clients[i].timeout = 0;                        
            *acc_id = macc_id;
        }
        // new account .. add to clients account
        else
        {
            j = client_account_by_free(i);

            if(j != -1)
            {                  
                // Save Host+Port from client (VIA)                
                pj_ansi_snprintf(oGLB.clients[i].addr, sizeof(oGLB.clients[i].addr), "%s", client);
                //pj_strdup_with_null(oGLB.pool, &oGLB.clients[i].addr, &cur_client);
                oGLB.clients[i].account_id[j] = macc_id;
                oGLB.clients[i].timeout = 0;  
                oGLB.clients[i].contact_count = 0;

                // Save all contacts from client
                client_contact = (pjsip_contact_hdr*) pjsip_msg_find_hdr(rdata->msg_info.msg, PJSIP_H_CONTACT, NULL);
                if(client_contact)
                {                   
                    pjsip_contact_hdr *hcnext = client_contact;
                    int c = oGLB.clients[i].contact_count;

                    while(hcnext && (c < MAX_CONTACTS))
                    {
                        if(hcnext->type == PJSIP_H_CONTACT)
                        {
                            char tmp[MAX_URI_LENGTH];                        
                            pjsip_sip_uri *contact_uri;

                            pj_memset(oGLB.clients[i].contacts[c], 0, sizeof(oGLB.clients[i].contacts[c]));
                            contact_uri = (pjsip_sip_uri*) pjsip_uri_get_uri(client_contact->uri);
                            pjsip_uri_print(PJSIP_URI_IN_CONTACT_HDR, contact_uri, tmp, sizeof(tmp));

                            if(strlen(tmp))
                                // contact header found!
                                pj_ansi_snprintf(oGLB.clients[i].contacts[c], sizeof(oGLB.clients[i].contacts[c]), "%s", tmp);
                            else {
                                // no contact header? generate from VIA
                                if(c == 0)
                                    pj_ansi_snprintf(oGLB.clients[i].contacts[c], sizeof(oGLB.clients[i].contacts[c]), "sip:%s", client);
                            }
                            
                            // add parameters to address
                            if(c == 0) {
                                strncpy(oGLB.clients[i].contact_param, tmp, sizeof(oGLB.clients[i].contact_param));
                                sip_strip_host(oGLB.clients[i].contact_param);                    
                                get_sep(oGLB.clients[i].contact_param, tmp, sizeof(tmp), ';');                            
                            }

                            oGLB.clients[i].contact_count++;
                            c = oGLB.clients[i].contact_count;
                            hcnext = hcnext->next;
                        }
                        else
                            break;                        
                    }
                }
                
                add = 2;
                *acc_id = macc_id;
                app_msg(THIS_FILE, "Client %s authorized for account %s", client, oSET.accounts[macc_id].name);
            }
            else
            {                
                add = 0;                
            }
        }

        // Max. number reached?
        if(!add)
        {
            app_err(THIS_FILE, "Unable to authorize client %s! Max. %d client(s) allowed!", client, oSET.voip.max_clients);
            statuscode = 487;
        }
    }

    // Return statuscode for response
    return statuscode;
}


// ------------------------------------------------------------------------------------------------------------
//
//  AUTHORIZATION (DTMFBOX -> REGISTRAR)
//
// ------------------------------------------------------------------------------------------------------------

// Client REGISTER Callback
static void voip_client_regcb(struct pjsip_regc_cbparam *param)
{    
    int id = (int)param->token;
    pj_str_t registrar = pj_str(oSET.accounts[id].registrar);
    
    //DTMFBOX_LOCK();

    //oSET.accounts[id].status = 0;                                       // set account to pending    
    //if(oSET.voip.keep_alive != 0) {
      //  pjsip_endpt_cancel_timer(oGLB.endpt, &oGLB.keep_alive_timer);   // Stop keep-alive callback        
    //}    

    // We got rdata!
    if(param->rdata != NULL)
    {
        // Show sip RX-REGISTER message
        app_dbg2(THIS_FILE, "SIP (RX-REGISTER):\n\n%s", param->rdata->msg_info.msg_buf);

        if (param->status!=PJ_SUCCESS) {
            app_err(THIS_FILE, "SIP registration error (%s), status=%d!", registrar, param->status);
            app_err(THIS_FILE, "%d - %.*s", param->code, param->reason.slen, param->reason.ptr);            
            oSET.accounts[id].status_nextreg = -1;
            oSET.accounts[id].status_failcnt++;
            oSET.accounts[id].status = 2;  // Error!
            pj_ansi_snprintf(oSET.accounts[id].status_reason, sizeof(oSET.accounts[id].status_reason), "(%d) - %.*s - Try: %d", param->code, param->reason.slen, param->reason.ptr, oSET.accounts[id].status_failcnt);
        }
        else if(param->code < 0 || param->code >= 300)
        {
            app_err(THIS_FILE, "SIP registration failed (%s), status=%d!", registrar, param->status);
            app_err(THIS_FILE, "%d - %.*s", param->code, param->reason.slen, param->reason.ptr);            
            oSET.accounts[id].status_nextreg = -1;
            oSET.accounts[id].status_failcnt++;
            oSET.accounts[id].status = 2;  // Error!
            pj_ansi_snprintf(oSET.accounts[id].status_reason, sizeof(oSET.accounts[id].status_reason), "(%d) - %.*s - Try: %d", param->code, param->reason.slen, param->reason.ptr, oSET.accounts[id].status_failcnt);
        }
        else if(PJSIP_IS_STATUS_IN_CLASS(param->code, 200))
        {
            if(param->expiration < 1)
            {
                // Unregister successful!
                app_msg(THIS_FILE, "SIP unregistration successful (%s), status=%d!", registrar, param->status);                
                oSET.accounts[id].status_nextreg = 0;
                oSET.accounts[id].status = 0; // Pending
                pj_ansi_snprintf(oSET.accounts[id].status_reason, sizeof(oSET.accounts[id].status_reason), "(%d) - %.*s", param->code, param->reason.slen, param->reason.ptr);

                // destroy regc
                pjsip_regc_destroy(oSET.accounts[id].regc);
                oSET.accounts[id].regc = NULL;
            }
            else
            {
                // Registration successful!
                app_msg(THIS_FILE, "SIP registration successful (%s), status=%d", registrar, param->status);
                app_msg(THIS_FILE, "Next registration in %d sec", param->expiration);                
                oSET.accounts[id].status_nextreg = param->expiration; //+ oSET.voip.keep_alive;
                oSET.accounts[id].status_failcnt = 0;
                oSET.accounts[id].status = 1;
                pj_ansi_snprintf(oSET.accounts[id].status_reason, sizeof(oSET.accounts[id].status_reason), "(%d) - %.*s", param->code, param->reason.slen, param->reason.ptr);
            }
        }
        else
        {
            app_msg(THIS_FILE, "SIP registration updated (%s), status=%d", registrar, param->status);
            pj_ansi_snprintf(oSET.accounts[id].status_reason, sizeof(oSET.accounts[id].status_reason), "(%d) - %.*s", param->code, param->reason.slen, param->reason.ptr);
        }
    }
    // No rdata??
    else
    {
        app_err(THIS_FILE, "SIP (RX-REGISTER): %.*s", param->reason.slen, param->reason.ptr);        
        oSET.accounts[id].status_nextreg = -1;
        oSET.accounts[id].status_failcnt++;
        oSET.accounts[id].status = 2;  // Error!
        pj_ansi_snprintf(oSET.accounts[id].status_reason, sizeof(oSET.accounts[id].status_reason), "(%d) - %.*s - Try: %d", param->code, param->reason.slen, param->reason.ptr, oSET.accounts[id].status_failcnt);

        // destroy regc
        pjsip_regc_destroy(oSET.accounts[id].regc);
        oSET.accounts[id].regc = NULL;
    }

    // destroy regc on failure..
    if(param->status!=PJ_SUCCESS && oSET.accounts[id].regc != NULL)
    {
        pjsip_regc_destroy(oSET.accounts[id].regc);
        oSET.accounts[id].regc = NULL;
    }

    /*
    // Send keep-Alive in the next 15 seconds
    if(oSET.voip.keep_alive != 0)
    {        
        pj_time_val timeout;                            
        timeout.sec = 15;                            
        timeout.msec = 0;
        pjsip_endpt_schedule_timer(oGLB.endpt, &oGLB.keep_alive_timer, &timeout);  
    }
    */

    //DTMFBOX_UNLOCK();
}

#endif

// Initialize registration (register at provider)
// if acc_id = -1, all accounts get registered.
int voip_client_register_init(int acc_id)
{    
    #if DTMFBOX_USE_VOIP
    
    char registrar_buf[MAX_URI_LENGTH];
    char fromuri_buf[MAX_URI_LENGTH];    
    char contact_buf[1024];
    pj_pool_t *pool;
    pjsip_tx_data *tx_data = 0;
    pj_str_t registrar;    
    pj_str_t fromuri;
    pj_str_t contact;
    pj_status_t status;
    #endif

    char localuri_buf[MAX_URI_LENGTH];
    pj_str_t localuri;
    int start = 0;
    int end = MAX_ACCOUNTS;
    int i = 0;
        
    if(acc_id >= 0 && acc_id < MAX_ACCOUNTS)
    {
        start = acc_id;
        end = acc_id+1;
    }

    for(i=start; i<end; i++)
    {
         // set to pending..
        oSET.accounts[i].status = 0;
        oSET.accounts[i].status_nextreg = 0;

        if(strlen(oSET.accounts[i].number) && oSET.accounts[i].active == 1)
        {
            // no username? then set number as username
            if(strlen(oSET.accounts[i].username) == 0)
                pj_ansi_snprintf(oSET.accounts[i].username, sizeof(oSET.accounts[i].username), "%s", oSET.accounts[i].number);

            // create local id, if not defined
            if(strlen(oSET.accounts[i].id))
            {
                // user-defined (from settings), with host and without (create one)
                if(strchr(oSET.accounts[i].id, (int)'@') != NULL)
                {
                    pj_ansi_snprintf(localuri_buf, sizeof(localuri_buf), "<sip:%s>", oSET.accounts[i].id);
                    localuri = pj_str(localuri_buf);
                }
                else
                {                    
                    pj_ansi_snprintf(localuri_buf, sizeof(localuri_buf), "<sip:%s@%s:%d>", oSET.accounts[i].id, oGLB.ip_addr_global, oGLB.udp_port_global);
                    localuri = pj_str(localuri_buf);
                    pj_ansi_snprintf(oSET.accounts[i].id, sizeof(oSET.accounts[i].id), "%s@%s:%d", oSET.accounts[i].id, oGLB.ip_addr_global, oGLB.udp_port_global);
                }
            }
            else
            {
                // create one (username@resolvedip:port)
                pj_ansi_snprintf(localuri_buf, sizeof(localuri_buf), "<sip:%s@%s:%d>", oSET.accounts[i].username, oGLB.ip_addr_global, oGLB.udp_port_global);
                localuri = pj_str(localuri_buf);
                pj_ansi_snprintf(oSET.accounts[i].id, sizeof(oSET.accounts[i].id), "%s@%s:%d", oSET.accounts[i].username, oGLB.ip_addr_global, oGLB.udp_port_global);
            }
           
            #if DTMFBOX_USE_VOIP

                // Register at SIP-Registrar
                if(strlen(oSET.accounts[i].realm) &&
                   strlen(oSET.accounts[i].registrar) &&
                   oSET.accounts[i].controller.device == VOIP &&
                   oSET.accounts[i].active &&
                   !oSET.accounts[i].do_not_register &&
                   oSET.voip_active)
                {
                    // set credential
                    oSET.accounts[i].cred.realm     = pj_str(oSET.accounts[i].realm);
                    oSET.accounts[i].cred.username  = pj_str(oSET.accounts[i].username);
                    oSET.accounts[i].cred.data_type = PJSIP_CRED_DATA_PLAIN_PASSWD;
                    oSET.accounts[i].cred.scheme    = pj_str("digest");
                    oSET.accounts[i].cred.data      = pj_str(oSET.accounts[i].password);

                    // registrar
                    pj_ansi_snprintf(registrar_buf, sizeof(registrar_buf), "<sip:%s>", oSET.accounts[i].registrar);
                    registrar = pj_str(registrar_buf);

                    // from-uri
                    pj_ansi_snprintf(fromuri_buf, sizeof(fromuri_buf), "\"%s\" <sip:%s>", oSET.accounts[i].name, oSET.accounts[i].id);
                    fromuri = pj_str(fromuri_buf);
                    
                    // create contact if not defined (username@localip:udp_port)            
                    pool = pj_pool_create_on_buf(NULL, contact_buf, sizeof(contact_buf));
                    status = voip_create_uac_contact(pool, &contact, i, &registrar);

                    // create regc and init (i is the account-id in the callback, token)
                    pjsip_regc_create(oGLB.endpt, (void*)i, &voip_client_regcb, &oSET.accounts[i].regc);

                    // Init regc
                    status = pjsip_regc_init(oSET.accounts[i].regc, &registrar, &fromuri, &localuri, 1, &contact, (pj_uint32_t)oSET.voip.register_interval);
                    if (status != PJ_SUCCESS) {
                        app_err(THIS_FILE, "(%d) Client registration initialization error", status);
                        oSET.accounts[i].status = 2;  // Error!
                        return status;
                    }

                    // set credentials
                    status = pjsip_regc_set_credentials(oSET.accounts[i].regc, 1, &oSET.accounts[i].cred);
                    if (status != PJ_SUCCESS) {
                        app_err(THIS_FILE, "(%d) Client authorization error", status);
                        oSET.accounts[i].status = 2;  // Error!
                        return status;
                    }

                    // Outbound-proxy?
                    if(strlen(oSET.accounts[i].proxy) > 0)
                    {                        
                        pjsip_route_hdr *h;                        
                        char uri[MAX_URI_LENGTH];
                        const pj_str_t hname = { "Route", 5 };
                        pj_str_t tmp;

                        pj_ansi_snprintf(uri, sizeof(uri), "sip:%s;lr", oSET.accounts[i].proxy);
                        uri[strlen(uri)] = '\0';
	                    pj_strdup2_with_null(oGLB.pool, &tmp, (const char*)&uri);

                        pj_list_init(&oSET.accounts[i].route_set);
                        h = (pjsip_route_hdr*)pjsip_parse_hdr(oGLB.pool, &hname, tmp.ptr, tmp.slen, NULL);
                        if(h != NULL)
                        {
                            pj_list_init(&oSET.accounts[i].route_set);
                            pj_list_push_back(&oSET.accounts[i].route_set, h);
                            
                            if (!pj_list_empty(&oSET.accounts[i].route_set)) {
                                pjsip_regc_set_route_set(oSET.accounts[i].regc, &oSET.accounts[i].route_set );
                            }
                        }
                    }

                    // Add user-agent header ;)
                    {                       
                       pjsip_hdr hdr_list;
                       pj_list_init(&hdr_list);
                       pj_list_push_back(&hdr_list, (pjsip_hdr*)oGLB.useragent_hdr);
                       pjsip_regc_add_headers(oSET.accounts[i].regc, &hdr_list);
                    }

                    // Register callback and send registration
                    pjsip_regc_register(oSET.accounts[i].regc, PJ_TRUE, &tx_data);
                    pjsip_regc_send(oSET.accounts[i].regc, tx_data);

                    // Show sip TX-REGISTER message
                    app_dbg2(THIS_FILE, "SIP (TX-REGISTER):\n\n%s", tx_data->buf);
                }

            #endif
        }
    }

    return 0;
}

// Deinit registration (unregister at provider)
// if acc_id = -1, all accounts get unregistered.
int voip_client_register_deinit(int acc_id)
{
    #if DTMFBOX_USE_VOIP
    
    pjsip_tx_data *tx_data = 0;
    //int i = 0;
    int j = 0;
    int start = 0;
    int end = MAX_ACCOUNTS;    
    pj_status_t status = PJ_SUCCESS;

    if(acc_id >= 0 && acc_id < MAX_ACCOUNTS)
    {
        start = acc_id;
        end = acc_id+1;
    }

    for(j=start; j<end; j++)
    {
        if(oSET.accounts[j].regc != NULL)
        {
            // Create unregister message
            status = pjsip_regc_unregister(oSET.accounts[j].regc, &tx_data);

            if(status == PJ_SUCCESS)
            {                
                // Send unregister messager
                status = pjsip_regc_send(oSET.accounts[j].regc, tx_data);      // send unregister
                if(status != PJ_SUCCESS) {
                    app_err(THIS_FILE, "(%d) Unable to send unregister (%s)", status, oSET.accounts[j].registrar);                    
                }
                else
                {
                    app_msg(THIS_FILE, "Unregistration successfull (%s)", oSET.accounts[j].registrar);
                }

                status = pjsip_regc_destroy(oSET.accounts[j].regc);            // destroy regc structure
                if(status == PJ_SUCCESS)
                {                    
                    oSET.accounts[j].regc = NULL;
                    oSET.accounts[j].status = 0;                               // set pending..                    
                }
            } else {
                oSET.accounts[j].status = 0;                               // set pending..                    
                app_err(THIS_FILE, "(%d) Unable to unregister (%s)", status, oSET.accounts[j].registrar);
            }
        }
    }
    
    return status;

    #else
    PJ_UNUSED_ARG(acc_id);
    return 0;
    #endif

}
